program DegugCoreTest;

{$APPTYPE CONSOLE}

uses
  Windows,
  SysUtils,
  FWDebugerCore in 'FWDebugerCore.pas';

type
  TTestDebugger = class
  private
    FCore: TFWDebugerCore;
  protected
    procedure OnCreateProcess(Sender: TObject; ThreadIndex: Integer;
      Data: TCreateProcessDebugInfo);
    procedure OnLoadDll(Sender: TObject; ThreadIndex: Integer;
      Data: TLoadDLLDebugInfo);
    procedure OnDebugString(Sender: TObject; ThreadIndex: Integer;
      Data: TOutputDebugStringInfo);

    procedure OnBreakPoint(Sender: TObject; ThreadIndex: Integer;
      ExceptionRecord: Windows.TExceptionRecord; BreakPointIndex: Integer;
      var ReleaseBreakpoint: Boolean);
    procedure OnHardwareBreakPoint(Sender: TObject; ThreadIndex: Integer;
      ExceptionRecord: Windows.TExceptionRecord; BreakPointIndex: THWBPIndex;
      var ReleaseBreakpoint: Boolean);
    procedure OnUnknownBreakPoint(Sender: TObject; ThreadIndex: Integer;
      ExceptionRecord: Windows.TExceptionRecord);
    procedure OnUnknownException(Sender: TObject; ThreadIndex: Integer;
      ExceptionRecord: Windows.TExceptionRecord);
  public
    constructor Create(const Path: string);
    destructor Destroy; override;
    procedure RunTest;
    procedure HideDebugger(hProcess: THandle);
  end;

{ TTestDebugger }

constructor TTestDebugger.Create(const Path: string);
begin
  FCore := TFWDebugerCore.Create;
  if not FCore.DebugNewProcess(Path, True) then
    RaiseLastOSError;
  FCore.OnCreateProcess := OnCreateProcess;
  FCore.OnLoadDll := OnLoadDll;
  FCore.OnDebugString := OnDebugString;
  FCore.OnBreakPoint := OnBreakPoint;
  FCore.OnHardwareBreakpoint := OnHardwareBreakpoint;
  FCore.OnUnknownBreakPoint := OnUnknownBreakPoint;
  FCore.OnUnknownException := OnUnknownException;
end;

destructor TTestDebugger.Destroy;
begin
  FCore.Free;
  inherited;
end;

const
  STATUS_SUCCESS = 0;
 
type
  NTStatus = Cardinal;
 
  UNICODE_STRING = record
    Length: Word;
    MaximumLength: Word;
    Buffer: PWideChar;
  end;
 
  RTL_DRIVE_LETTER_CURDIR = record
    Flags: Word;
    Length: Word;
    TimeStamp: Cardinal;
    DosPath: UNICODE_STRING;
  end;
 
  PRTL_USER_PROCESS_PARAMETERS = ^RTL_USER_PROCESS_PARAMETERS;
  RTL_USER_PROCESS_PARAMETERS = record
    MaximumLength: Cardinal;
    Length: Cardinal;
    Flags: Cardinal;
    DebugFlags: Cardinal;
    ConsoleHandle: Pointer;
    ConsoleFlags: Cardinal;
    StdInputHandle: Cardinal;
    StdOutputHandle: Cardinal;
    StdErrorHandle: Cardinal;
    CurrentDirectoryPath: UNICODE_STRING;
    CurrentDirectoryHandle: Cardinal;
    DllPath: UNICODE_STRING;
    ImagePathName: UNICODE_STRING;
    CommandLine: UNICODE_STRING;
    Environment: Pointer;
    StartingPositionLeft: Cardinal;
    StartingPositionTop: Cardinal;
    Width: Cardinal;
    Height: Cardinal;
    CharWidth: Cardinal;
    CharHeight: Cardinal;
    ConsoleTextAttributes: Cardinal;
    WindowFlags: Cardinal;
    ShowWindowFlags: Cardinal;
    WindowTitle: UNICODE_STRING;
    DesktopName: UNICODE_STRING;
    ShellInfo: UNICODE_STRING;
    RuntimeData: UNICODE_STRING;
    DLCurrentDirectory: array [0..31] of RTL_DRIVE_LETTER_CURDIR
  end;
 
  PPEB=^PEB;
  PEB = record
    InheritedAddressSpace: Boolean;
    ReadImageFileExecOptions: Boolean;
    BeingDebugged: Boolean;
    Spare: Boolean;
    Mutant: Cardinal;
    ImageBaseAddress: Pointer;
    LoaderData: Pointer; //PPEB_LDR_DATA;
    ProcessParameters: PRTL_USER_PROCESS_PARAMETERS;
    SubSystemData: Pointer;
    ProcessHeap: Pointer;
    FastPebLock: Pointer;
    FastPebLockRoutine: Pointer;
    FastPebUnlockRoutine: Pointer;
    EnvironmentUpdateCount: Cardinal;
    KernelCallbackTable: PPointer;
    EventLogSection: Pointer;
    EventLog: Pointer;
    FreeList: Pointer; //PPEB_FREE_BLOCK;
    TlsExpansionCounter: Cardinal;
    TlsBitmap: Pointer;
    TlsBitmapBits: array[0..1] of Cardinal;
    ReadOnlySharedMemoryBase: Pointer;
    ReadOnlySharedMemoryHeap: Pointer;
    ReadOnlyStaticServerData: PPointer;
    AnsiCodePageData: Pointer;
    OemCodePageData: Pointer;
    UnicodeCaseTableData: Pointer;
    NumberOfProcessors: Cardinal;
    NtGlobalFlag: Cardinal;
    Spare2: array[0..3] of Byte;
    CriticalSectionTimeout: LARGE_INTEGER;
    HeapSegmentReserve: Cardinal;
    HeapSegmentCommit: Cardinal;
    HeapDeCommitTotalFreeThreshold: Cardinal;
    HeapDeCommitFreeBlockThreshold: Cardinal;
    NumberOfHeaps: Cardinal;
    MaximumNumberOfHeaps: Cardinal;
    ProcessHeaps: Pointer;
    GdiSharedHandleTable: Pointer;
    ProcessStarterHelper: Pointer;
    GdiDCAttributeList: Pointer;
    LoaderLock: Pointer;
    OSMajorVersion: Cardinal;
    OSMinorVersion: Cardinal;
    OSBuildNumber: Cardinal;
    OSPlatformId: Cardinal;
    ImageSubSystem: Cardinal;
    ImageSubSystemMajorVersion: Cardinal;
    ImageSubSystemMinorVersion: Cardinal;
    GdiHandleBuffer: array [0..33] of Cardinal;
    PostProcessInitRoutine: Cardinal;
    TlsExpansionBitmap: Cardinal;
    TlsExpansionBitmapBits: array [0..127] of Byte;
    SessionId: Cardinal;
  end;
 
  PPROCESS_BASIC_INFORAMTION=^PROCESS_BASIC_INFORMATION;
  PROCESS_BASIC_INFORMATION = record
    ExitStatus: NTStatus;
    PebBaseAddress: PPEB;
    AffinityMask: PCardinal;
    BasePriority: Cardinal;
    uUniqueProcessId: Cardinal;
    uInheritedFromUniqueProcessId: Cardinal;
  end;
 
function NtQueryInformationProcess(ProcessHandle: Cardinal;
  ProcessInformationClass: Integer;
  ProcessInformation: Pointer;
  ProcessInformationLength: Cardinal;
  ReturnLength: PCardinal): NTStatus; stdcall; external 'ntdll.dll';

procedure TTestDebugger.HideDebugger(hProcess: THandle);
var
  pProcBasicInfo: PROCESS_BASIC_INFORMATION;
  pPeb: PEB;
  ReturnLength: DWORD;
begin
  if NtQueryInformationProcess(hProcess, 0,
    @pProcBasicInfo, SizeOf(PROCESS_BASIC_INFORMATION),
    @ReturnLength) <> STATUS_SUCCESS then
    RaiseLastOSError;
  if not ReadProcessMemory(hProcess, pProcBasicInfo.PebBaseAddress,
    @pPeb, SizeOf(PEB), ReturnLength) then
    RaiseLastOSError;
  pPeb.BeingDebugged := False;
  if not WriteProcessMemory(hProcess, pProcBasicInfo.PebBaseAddress,
    @pPeb, SizeOf(PEB), ReturnLength) then
    RaiseLastOSError;
end;

procedure TTestDebugger.OnBreakPoint(Sender: TObject; ThreadIndex: Integer;
  ExceptionRecord: Windows.TExceptionRecord; BreakPointIndex: Integer;
  var ReleaseBreakpoint: Boolean);
begin
  if ExceptionRecord.ExceptionAddress =
    Pointer(FCore.DebugProcessData.EntryPoint) then
  begin
    Writeln;
    Writeln(Format('!!! --> Process Entry Point found. Address: %p',
      [Pointer(FCore.DebugProcessData.EntryPoint)]));
    Writeln;

    HideDebugger(FCore.DebugProcessData.AttachedProcessHandle);

    FCore.SetHardwareBreakpoint(ThreadIndex, Pointer($467840), hsByte,
      hmExecute, 0, 'wait JE');
  end
  else
  begin
    Writeln;
    Writeln(Format('!!! --> BreakPoint at addr 0X%p - "%s"',
      [ExceptionRecord.ExceptionAddress,
      FCore.BreakpointItem(BreakPointIndex).Description]));
    Writeln;
  end;
end;

procedure TTestDebugger.OnCreateProcess(Sender: TObject; ThreadIndex: Integer;
  Data: TCreateProcessDebugInfo);
var
  T: TThreadData;
begin
  T := FCore.GetThreadData(ThreadIndex);
  Writeln(Format('CreateThread ID: %d', [T.ThreadID, T.ThreadHandle]));
  Writeln(Format('ProcessStart ID: %d', [FCore.DebugProcessData.ProcessID]));
end;

procedure TTestDebugger.OnDebugString(Sender: TObject; ThreadIndex: Integer;
  Data: TOutputDebugStringInfo);
begin
  if Data.fUnicode = 1 then
    Writeln('DebugString: ' + FCore.ReadStringW(Data.lpDebugStringData, Data.nDebugStringLength))
  else
    Writeln('DebugString: ' + FCore.ReadStringA(Data.lpDebugStringData, Data.nDebugStringLength));
end;

procedure TTestDebugger.OnHardwareBreakPoint(Sender: TObject;
  ThreadIndex: Integer; ExceptionRecord: Windows.TExceptionRecord;
  BreakPointIndex: THWBPIndex; var ReleaseBreakpoint: Boolean);
var
  ThreadData: TThreadData;
begin
  Writeln;
  ThreadData := FCore.GetThreadData(ThreadIndex);
  Writeln(Format('!!! --> Hardware BreakPoint at addr 0X%p - "%s"',
    [ExceptionRecord.ExceptionAddress,
    ThreadData.Breakpoint.Description[BreakPointIndex]]));
  FCore.SetFlag(ThreadIndex, EFLAGS_ZF, True);
  Writeln;
end;

procedure TTestDebugger.OnLoadDll(Sender: TObject; ThreadIndex: Integer;
  Data: TLoadDLLDebugInfo);
const
  FormatStrKnownDLL = 'Load Dll at instance %p handle %d "%s"';
  FormatStrUnknownDLL = 'Load unknown Dll at instance %p handle %d';
var
  DllName: AnsiString;
  IsUnicodeData: Boolean;
begin
  IsUnicodeData := Data.fUnicode = 1;
  DllName := FCore.GetDllName(Data.lpImageName, Data.lpBaseOfDll, IsUnicodeData);
  if DllName <> '' then
  begin
    if IsUnicodeData then
      Writeln(Format(FormatStrKnownDLL, [Data.lpBaseOfDll,
        Data.hFile, PWideChar(@DllName[1])]))
    else
      Writeln(Format(FormatStrKnownDLL, [Data.lpBaseOfDll,
        Data.hFile, PAnsiChar(@DllName[1])]));
  end
  else
    Writeln(Format(FormatStrUnknownDLL, [Data.lpBaseOfDll, Data.hFile]));
end;

procedure TTestDebugger.OnUnknownBreakPoint(Sender: TObject;
  ThreadIndex: Integer; ExceptionRecord: Windows.TExceptionRecord);
begin
  Writeln;
  Writeln(Format('!!! --> Unknown breakpoint at addr 0X%p',
    [ExceptionRecord.ExceptionAddress]));
  Writeln;
end;

procedure TTestDebugger.OnUnknownException(Sender: TObject;
  ThreadIndex: Integer; ExceptionRecord: Windows.TExceptionRecord);
begin
  Writeln;
  Writeln(Format('!!! --> Unknown exception code %p at addr 0X%p',
    [
      Pointer(ExceptionRecord.ExceptionCode),
      ExceptionRecord.ExceptionAddress
    ]));
  Writeln;
end;

procedure TTestDebugger.RunTest;
begin
  FCore.RunMainLoop;
end;

var
  Tst: TTestDebugger;

begin
  try
    Tst := TTestDebugger.Create(ExtractFilePath(ParamStr(0)) +  'crackme\crackme.exe');
    try
      Tst.RunTest;
    finally
      Tst.Free;
    end;
  except
    on E: Exception do
      writeln(E.ClassName + ': ' + E.Message);
  end;
  Writeln('DONE');
  Readln;
end.
